# Copyright 2016 - Alcatel-Lucent
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import EntityCategory
from vitrage.common.constants import GraphAction
from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources.neutron.network import NEUTRON_NETWORK_DATASOURCE
from vitrage.datasources.neutron.properties import NetworkProperties\
    as NetworkProps
from vitrage.datasources.resource_transformer_base import \
    ResourceTransformerBase
from vitrage.datasources import transformer_base as tbase
from vitrage.datasources.transformer_base import extract_field_value
import vitrage.graph.utils as graph_utils


class NetworkTransformer(ResourceTransformerBase):

    UPDATE_ID_PROPERTY = {
        'network.create.end': (NetworkProps.NETWORK, NetworkProps.ID),
        'network.update.end': (NetworkProps.NETWORK, NetworkProps.ID),
        'network.delete.end': ('network_id',),
        None: (NetworkProps.ID,)
    }

    # graph actions which need to refer them differently
    GRAPH_ACTION_MAPPING = {
        'network.delete.end': GraphAction.DELETE_ENTITY,
    }

    def _create_snapshot_entity_vertex(self, entity_event):

        name = entity_event[NetworkProps.NAME]
        entity_id = entity_event[NetworkProps.ID]
        state = entity_event[NetworkProps.STATUS]
        update_timestamp = entity_event[NetworkProps.UPDATED_AT]
        project_id = entity_event.get(NetworkProps.TENANT_ID, None)

        return self._create_vertex(entity_event,
                                   name,
                                   entity_id,
                                   state,
                                   update_timestamp,
                                   project_id)

    def _create_update_entity_vertex(self, entity_event):

        event_type = entity_event[DSProps.EVENT_TYPE]
        name = extract_field_value(entity_event, NetworkProps.NETWORK,
                                   NetworkProps.NAME)
        state = extract_field_value(entity_event, NetworkProps.NETWORK,
                                    NetworkProps.STATUS)
        update_timestamp = \
            extract_field_value(entity_event, NetworkProps.NETWORK,
                                NetworkProps.UPDATED_AT)
        entity_id = extract_field_value(entity_event,
                                        *self.UPDATE_ID_PROPERTY[event_type])
        project_id = extract_field_value(entity_event, NetworkProps.NETWORK,
                                         NetworkProps.TENANT_ID)

        return self._create_vertex(entity_event,
                                   name,
                                   entity_id,
                                   state,
                                   update_timestamp,
                                   project_id)

    def _create_vertex(self,
                       entity_event,
                       name,
                       entity_id,
                       state,
                       update_timestamp,
                       project_id):

        metadata = {
            VProps.NAME: name,
            VProps.PROJECT_ID: project_id,
        }

        vitrage_sample_timestamp = entity_event[DSProps.SAMPLE_DATE]

        return graph_utils.create_vertex(
            self._create_entity_key(entity_event),
            vitrage_category=EntityCategory.RESOURCE,
            vitrage_type=NEUTRON_NETWORK_DATASOURCE,
            vitrage_sample_timestamp=vitrage_sample_timestamp,
            entity_id=entity_id,
            entity_state=state,
            update_timestamp=update_timestamp,
            metadata=metadata)

    def _create_entity_key(self, entity_event):
        event_type = entity_event.get(DSProps.EVENT_TYPE, None)
        network_id = extract_field_value(entity_event,
                                         *self.UPDATE_ID_PROPERTY[event_type])

        key_fields = self._key_values(NEUTRON_NETWORK_DATASOURCE, network_id)
        return tbase.build_key(key_fields)

    def get_vitrage_type(self):
        return NEUTRON_NETWORK_DATASOURCE



# Copyright 2016 - Alcatel-Lucent
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re
from vitrage.common.constants import EdgeProperties as EConst
from vitrage.common.constants import VertexProperties as VConst
from vitrage.graph.driver.elements import Edge
from vitrage.graph.driver.elements import Vertex


def create_vertex(vitrage_id,
                  vitrage_category=None,
                  vitrage_type=None,
                  vitrage_sample_timestamp=None,
                  vitrage_is_deleted=False,
                  vitrage_is_placeholder=False,
                  entity_id=None,
                  entity_state=None,
                  update_timestamp=None,
                  project_id=None,
                  vitrage_resource_project_id=None,
                  metadata=None,
                  datasource_name=None):
    """A builder to create a vertex

    :param vitrage_id:
    :type vitrage_id: str
    :param entity_id:
    :type entity_id: str
    :param vitrage_category:
    :type vitrage_category: str
    :param vitrage_type:
    :type vitrage_type: str
    :param entity_state:
    :type entity_state: str
    :param vitrage_is_deleted:
    :type vitrage_is_deleted: boolean
    :param update_timestamp:
    :type update_timestamp: str
    :param vitrage_sample_timestamp:
    :type vitrage_sample_timestamp: str
    :param metadata:
    :type metadata: dict
    :param vitrage_is_placeholder:
    :type vitrage_is_placeholder: boolean
    :param project_id:
    :type project_id: str
    :param datasource_name:
    :type datasource_name: str
    :return:
    :rtype: Vertex
    """

    properties = {
        VConst.ID: entity_id,
        VConst.STATE: entity_state,
        VConst.VITRAGE_TYPE: vitrage_type,
        VConst.VITRAGE_CATEGORY: vitrage_category,
        VConst.VITRAGE_IS_DELETED: vitrage_is_deleted,
        VConst.UPDATE_TIMESTAMP: update_timestamp,
        VConst.VITRAGE_SAMPLE_TIMESTAMP: vitrage_sample_timestamp,
        VConst.VITRAGE_IS_PLACEHOLDER: vitrage_is_placeholder,
        VConst.VITRAGE_ID: vitrage_id,
        VConst.PROJECT_ID: project_id,
        VConst.VITRAGE_RESOURCE_PROJECT_ID: vitrage_resource_project_id,
        VConst.VITRAGE_DATASOURCE_NAME: datasource_name,
    }
    if metadata:
        properties.update(metadata)
    properties = {k: v for k, v in properties.items() if v is not None}
    vertex = Vertex(vertex_id=vitrage_id, properties=properties)
    return vertex


def create_edge(source_id,
                target_id,
                relationship_type,
                vitrage_is_deleted=False,
                update_timestamp=None,
                metadata=None):
    """A builder to create an edge

    :param update_timestamp:
    :param source_id:
    :type source_id: str
    :param target_id:
    :type target_id: str
    :param relationship_type:
    :type relationship_type: str
    :param vitrage_is_deleted:
    :type vitrage_is_deleted: str
    :param metadata:
    :type metadata: dict
    :return:
    :rtype: Edge
    """
    properties = {
        EConst.UPDATE_TIMESTAMP: update_timestamp,
        EConst.VITRAGE_IS_DELETED: vitrage_is_deleted,
        EConst.RELATIONSHIP_TYPE: relationship_type,
    }
    if metadata:
        properties.update(metadata)
    properties = {k: v for k, v in properties.items() if v is not None}
    edge = Edge(source_id=source_id,
                target_id=target_id,
                label=relationship_type,
                properties=properties)
    return edge


def check_property_with_regex(key, regex, data):
    """Checks if the contents of data[key] matches the given regex

    :param: key: key to find in data
    :param: data: dict to search
    :type: data: dict
    :param: regex: regular expression to check against
    :type: regex: str
    :rtype: bool
    """
    value = data.get(key)
    if value is None:
        return False
    pattern = re.compile(regex)
    return pattern.match(value) is not None



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log

LOG = log.getLogger(__name__)


class TemplateSchemaFactory(object):

    DEFAULT_VERSION = '1'
    _schemas = dict()

    @classmethod
    def supported_versions(cls):
        return cls._schemas.keys()

    @classmethod
    def is_version_supported(cls, version):
        return version in cls._schemas

    @classmethod
    def template_schema(cls, version):
        if not version:
            version = cls.DEFAULT_VERSION
        template_schema = cls._schemas.get(version)
        return template_schema

    @classmethod
    def register_template_schema(cls, version, template_schema):
        cls._schemas[version] = template_schema
        LOG.debug('Registered template schema for version %s', version)



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import namedtuple

from oslo_config import cfg
from oslo_log import log
from oslo_utils import importutils as utils

from vitrage.common.constants import DatasourceAction
from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.datasources.alarm_driver_base import AlarmDriverBase
from vitrage.datasources.zabbix.properties import ZabbixProperties as ZProps
from vitrage.datasources.zabbix.properties import ZabbixTriggerStatus \
    as TriggerStatus
from vitrage.datasources.zabbix.properties import ZabbixTriggerValue \
    as TriggerValue
from vitrage.datasources.zabbix import ZABBIX_DATASOURCE
from vitrage.utils import file as file_utils

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class ZabbixDriver(AlarmDriverBase):
    ServiceKey = namedtuple('ServiceKey', ['hostname', 'triggerid'])
    conf_map = None

    def __init__(self):
        super(ZabbixDriver, self).__init__()
        if not ZabbixDriver.conf_map:
            ZabbixDriver.conf_map =\
                ZabbixDriver._configuration_mapping()
        self._client = None

    def zabbix_client_login(self):
        if not CONF.zabbix.user:
            LOG.warning('Zabbix user is not defined')
        if not CONF.zabbix.password:
            LOG.warning('Zabbix password is not defined')
        if not CONF.zabbix.url:
            LOG.warning('Zabbix url is not defined')

        try:
            if not self._client:
                self._client = utils.import_object(
                    'pyzabbix.ZabbixAPI',
                    CONF.zabbix.url)
            self._client.login(
                CONF.zabbix.user,
                CONF.zabbix.password)
        except Exception:
            LOG.exception('pyzabbix.ZabbixAPI error occurred.')
            self._client = None

    def _vitrage_type(self):
        return ZABBIX_DATASOURCE

    def _alarm_key(self, alarm):
        return self.ServiceKey(hostname=alarm[ZProps.RESOURCE_NAME],
                               triggerid=alarm[ZProps.TRIGGER_ID])

    def _get_alarms(self):
        self.zabbix_client_login()
        if not self._client:
            return []

        alarms = []
        valid_hosts = (host for host in
                       self._client.host.get(output=[ZProps.HOST])
                       if host[ZProps.HOST] in ZabbixDriver.conf_map)
        for host in valid_hosts:
            self._get_triggers_per_host(host, alarms)

        return alarms

    def _get_triggers_per_host(self, host, alarms):

        host_id = host[ZProps.HOST_ID]
        triggers = self._client.trigger.get(hostids=host_id,
                                            expandDescription=True)

        triggers_rawtexts = self._get_triggers_rawtexts(host_id)

        for trigger in triggers:
            trigger[ZProps.ZABBIX_RESOURCE_NAME] = host[ZProps.HOST]
            trigger_id = trigger[ZProps.TRIGGER_ID]
            trigger[ZProps.RAWTEXT] = triggers_rawtexts[trigger_id]
            alarms.append(trigger)

    def _get_triggers_rawtexts(self, host_id):

        output = [ZProps.TRIGGER_ID, ZProps.DESCRIPTION]
        triggers = self._client.trigger.get(hostids=host_id, output=output)

        return {trigger[ZProps.TRIGGER_ID]: trigger[ZProps.DESCRIPTION]
                for trigger in triggers}

    def _enrich_alarms(self, alarms):
        """Enrich zabbix alarm using zabbix configuration file

        converting Zabbix host name to Vitrage resource type and name

        :param alarms: Zabbix alarm
        :return: enriched alarm
        """

        for alarm in alarms:
            alarm[ZProps.VALUE] = self._get_value(alarm)

            zabbix_host = alarm[ZProps.ZABBIX_RESOURCE_NAME]
            vitrage_host = ZabbixDriver.conf_map[zabbix_host]
            alarm[ZProps.RESOURCE_TYPE] = vitrage_host[ZProps.RESOURCE_TYPE]
            alarm[ZProps.RESOURCE_NAME] = vitrage_host[ZProps.RESOURCE_NAME]

    def _is_erroneous(self, alarm):
        return alarm and \
            alarm[ZProps.VALUE] == TriggerValue.PROBLEM

    def _status_changed(self, new_alarm, old_alarm):

        if not (new_alarm and old_alarm):
            return False

        if new_alarm[ZProps.VALUE] != old_alarm[ZProps.VALUE]:
            return True

        if new_alarm[ZProps.VALUE] == TriggerValue.PROBLEM:
            priority_changed = \
                new_alarm[ZProps.PRIORITY] != old_alarm[ZProps.PRIORITY]
            description_changed = \
                new_alarm[ZProps.DESCRIPTION] != old_alarm[ZProps.DESCRIPTION]
            return priority_changed or description_changed

    def _is_valid(self, alarm):
        return alarm[ZProps.RESOURCE_TYPE] is not None and \
            alarm[ZProps.RESOURCE_NAME] is not None

    @staticmethod
    def _get_value(alarm):
        if alarm[ZProps.STATUS] == TriggerStatus.DISABLED:
            return TriggerValue.OK
        return alarm[ZProps.VALUE]

    @staticmethod
    def _configuration_mapping():
        try:
            zabbix_config_file = CONF.zabbix[DSOpts.CONFIG_FILE]
            zabbix_config = file_utils.load_yaml_file(zabbix_config_file)
            zabbix_config_elements = zabbix_config[ZABBIX_DATASOURCE]

            mappings = {}
            for element_config in zabbix_config_elements:
                mappings[element_config['zabbix_host']] = {
                    ZProps.RESOURCE_TYPE: element_config['type'],
                    ZProps.RESOURCE_NAME: element_config['name']
                }

            return mappings
        except Exception:
            LOG.exception('Failed in init.')
            return {}

    def enrich_event(self, event, event_type):
        event[DSProps.EVENT_TYPE] = event_type

        if ZabbixDriver.conf_map:
            zabbix_host = event[ZProps.HOST]
            event[ZProps.ZABBIX_RESOURCE_NAME] = zabbix_host
            v_resource = ZabbixDriver.conf_map[zabbix_host]
            event[ZProps.RESOURCE_NAME] = v_resource[ZProps.RESOURCE_NAME]
            event[ZProps.RESOURCE_TYPE] = v_resource[ZProps.RESOURCE_TYPE]

        return ZabbixDriver.make_pickleable([event], ZABBIX_DATASOURCE,
                                            DatasourceAction.UPDATE)[0]

    @staticmethod
    def get_event_types():
        return ['zabbix.alarm.ok', 'zabbix.alarm.problem']

    @staticmethod
    def should_delete_outdated_entities():
        return True
